###############################################################################
### Implantation de focntions de calcul de racines carrées dans
### des structures algébriques finies (corps finis, entiers modulo N), dans
### le cadre du cours d'Algorithmes Arithmétiques 2
### Auteur : J. Lavauzelle
### Date : octobre 2023
###############################################################################




# ============================================================================
### PREMIÈRE PARTIE : CORPS FINI


# Cas où q = 2^e
def sqrt_2e(x, q, e=None):
    if e is None:
        return x**(q//2)
    res = x
    for i in range(e-1):
        res = res**2
    return res
    # return x**(2**(e-1))

# Cas où q = 3 mod 4
def sqrt_q3mod4(x, q, s=None):
    if s is None:
        s = (q+1)//4
    return x**s


# Cas où q = 5 mod 8
def sqrt_q5mod8(x, q, s=None):
    if s is None:
        s = (q-5)//8
    v = (2*x)**s
    i = 2*x*v**2
    return x*v*(1-i)



# Algorithme de Cipolla
def cipolla(a):
    if a == 0:
        return a
    FF = a.parent()
    q = FF.cardinality()
    if q % 2 == 0:
        print("Warning: Cipolla algorithm only works for q odd. Switching algorithm...")
        return sqrt_2e(a, q)
    R = PolynomialRing(FF, "X")
    X = R.gen()
    y = FF.random_element()
    c = y**2 - a
    while c.is_square():
        y = FF.random_element()
        c = y**2 - a
    P = X**2 - c
    KK = R.quotient_ring(P)
    XX = KK.gen()
    res = (KK(y) + XX)**(int((q+1)/2))
    return res[0]




def sqrt_pk(a, p, k, verbose=False):
    Fp = GF(p)
    a0 = int(a)
    k0 = k
    if verbose:
        print("Entrée :\na = %d\np = %d\nk = %d\np^k = %d" % (a, p, k, p**k))

    # on élimine le cas où a est nul
    # (sinon on boucle infiniment quand on divise par p)
    if a == 0:
        if verbose:
            print("Cas où a = 0")
        return [ e * p**(k-k//2) for e in range(p**(k//2)) ]

    # on divise a par p tant que l'on peut, pour pouvoir
    # utiliser la remontée de Hensel
    cpt = 0
    while a0 % p == 0:
        a0 = a0//(p**2)
        k0 = k0 - 2
        cpt += 1
    if verbose:
        print("On a pu diviser a par p %d fois." % cpt)
        print("On se ramène à chercher les racines carrées de a0 = a/p^%d = %d modulo %d^%d = %d" % (cpt, a0, p, k0, p**k0))  
        
    # on calcule les racines carrées modulo p
    a0_p = Fp(a0)
    x1 = cipolla(a0_p)
    X = [ int(x1), p - int(x1) ]
    if verbose:
        print("L'algorithme de Cipolla nous donne les racines carrées suivantes de a0 mod p :")
        print(X)
    
    # remontée de Hensel
    if verbose:
        print("Remontée de Hensel :")
    R = []
    for x in X:
        if verbose:
            ell = ceil(log(p**k0, 10))+1
            print("Pour la racine x = %d" % x)
            print(" i |     t      |     x     ")
        for i in range(2, k0+1):
            z = (a0 - x**2)/(p**(i-1)) % p
            t = z * inverse_mod(2*x, p) % p
            x = x + t*p**(i-1)
            if verbose:
                print("%2d | %10d | %10d" % (i, t, x) )
        R.append(x)
        

    # reconstruction de toutes les solutions
    pcpt = p**cpt
    return [ pcpt * (r + e*p**k0) for r in R for e in range(pcpt) ]




def sqrt_2k(a, k):
    if k == 1:
        return [ a ]
    if k == 2:
        if a == 0:
            return [0, 2]
        elif a == 1:
            return [1, 3]
        else:
            return None
    if k == 3:
        if a == 0:
            return [0, 4]
        elif a == 1:
            return [1, 3, 5, 7]
        elif a == 4:
            return [2, 6]
        else:
            return None
    if a == 0:
        return [ e * 2**(k - k//2) for e in range(2**(k//2)) ]
    
    a0 = int(a)
    k0 = k
    res = 1
    cpt = 0
    while a0 % 2 == 0:
       a0 = a0//4
       k0 = k0 - 2
       res = res * 2
       cpt += 1
    if k0 <= 3:
        S = sqrt_2k(a0, k0)
        S = [ res * (x + e * 2**k0) for x in S for e in range(2**cpt) ]
        return S
    else:
        r = 1
        for kk in range(3, k0):
            if (a0 - r**2) % 2**(kk+1) != 0:
                r = r + 2**(kk-1) 
        if r < 2**(k0 - 1):
            S = [ r, 2**k0 - r, 2**(k0-1) + r , - r + 2**(k0-1) ]
        else:
            S = [ r, 2**k0 - r, r - 2**(k0-1) , 2**k0 - r + 2**(k0-1) ]
        S = [ res * (x + e * 2**k0) for x in S for e in range(2**cpt) ]
        return S



def crt_sqrt(L1, L2, N1, N2):
    _, u, v = xgcd(N1, N2)
    RES = [ ((N1 * u * x2 + N2 * v * x1) % (N1*N2)) for x1 in L1 for x2 in L2 ]
    return RES


def sqrt_modN(a, facto_N):
    m = len(facto_N)     
    L = []
    
    TMP = []
    for (p,k) in facto_N:
        if p == 2:
            TMP.append(sqrt_2k(a % 2**k, k))
        else:
            TMP.append(sqrt_pk(a % p**k, p, k))

    N = facto_N[0][0]**facto_N[0][1]
    RES = TMP[0]
    for i in range(1, m):
        new_N = facto_N[i][0]**facto_N[i][1]
        L1 = copy(RES)
        L2 = copy(TMP[i])
        RES = crt_sqrt(L1, L2, N, new_N)
        N = new_N * N
    return RES
        

