###############################################################################
### Implantation de focntions de calcul de racines carrées entières dans
### le cadre du cours d'Algorithmes Arithmétiques 2
### Auteur : J. Lavauzelle
### Date : octobre 2023
###############################################################################


# Calcule la décomposation de N en base 4
def decomposition_base_4(N):
    n = N
    res = []
    while n > 0:
        res.append(n % 4)
        n //= 4
    return res

# Calcule la racine carrée entière X de N (et son reste R
# par la méthode élémentaire (compl. O(log n) op.)
def integer_square_root(N):
    decomp = decomposition_base_4(N)
    d = len(decomp)
    R = 0
    X = 0
    for j in range(d-1, -1, -1):
        T = 4*R + decomp[j]
        if T > 4*X:
            R = T - 4*X - 1
            X = 2*X+1
        else:
            R = T
            X = 2*X
    return X, R




# Exécution de la méthode de Héron pour la racine carrée de N,
# jusqu'à une précision epsilon
def heron(N, epsilon, verbose=False):
    # on part avec x0 = N
    x = N
    new_x = (x + N/x)/2
    if verbose:
        print(new_x)
    while abs(new_x - x) > epsilon:
        x = new_x
        new_x = (x + N/x)/2
        if verbose:
            print(new_x)
    return new_x


# Calcule la racine carrée entière X de N (et son reste R
# par la méthode de Héron (compl. O(log n) op.)
def heron_int(N, verbose=False):
    epsilon = 0.0001
    N_f = float(N)            # on dit à sage de faire des opération flottantes
    x = heron(N_f, epsilon, verbose)
    x = floor(x)
    r = N - x*x
    return x, r


# Avec des flottants à précision arbitraire 
def heron_int_prec(N, prec=1000, verbose=False):
    RRR = RealField(prec)   # par ex: 1000 bits de précision !
    epsilon = RRR(1.0/10000)
    N_f = RRR(N)
    x = heron(N_f, epsilon, verbose)
    x = floor(x)
    r = N - x*x
    return x, r




